# Copyright (C) 2018 Avatao.com Innovative Learning Kft.
# All Rights Reserved. See LICENSE file for details.

from tfw import EventHandlerBase
from tfw.mixins import MonitorManagerMixin
from tfw.config.logs import logging
from .log_monitor import LogMonitor

LOG = logging.getLogger(__name__)


class LogMonitoringEventHandler(EventHandlerBase, MonitorManagerMixin):
    """
    Monitors the output of a supervisor process (stdout, stderr) and
    sends the results to the frontend.

    Accepts messages that have a data['command'] key specifying
    a command to be executed.

    The API of each command is documented in their respective handler.
    """
    def __init__(self, key, process_name, log_tail=0):
        super().__init__(key)
        self.process_name = process_name
        self.log_tail = log_tail
        MonitorManagerMixin.__init__(
            self,
            LogMonitor,
            self.process_name,
            self.log_tail
        )

        self.command_handlers = {
            'process_name': self.handle_process_name,
            'log_tail':     self.handle_log_tail
        }

    def handle_event(self, message):
        try:
            data = message['data']
            self.command_handlers[data['command']](data)
            self.reload_monitor()
        except KeyError:
            LOG.error('IGNORING MESSAGE: Invalid message received: %s', message)

    def handle_process_name(self, data):
        """
        Changes the monitored process.

        :param data: TFW message data containing 'value'
                     (name of the process to monitor)
        """
        self.set_monitor_args(data['value'], self.log_tail)

    def handle_log_tail(self, data):
        """
        Sets tail length of the log the monitor will send
        to the frontend (the monitor will send back the last
        'value' characters of the log).

        :param data: TFW message data containing 'value'
                     (new tail length)
        """
        self.set_monitor_args(self.process_name, data['value'])

    def cleanup(self):
        self.monitor.stop()
